// @flow
import _ from 'lodash';
import base from './base';
import {compose, lifecycle, withHandlers} from 'recompose';
import {connect} from 'react-redux';
import type {TFilterItem} from '../flow-typed';
import syncOnMount from './sync-on-mount';

const
	getCounter = (params: {
		counter: string,
		keyStatePath: string,
		state: Object,
	}) => _.toNumber(
		_.get(params.state, `${params.keyStatePath}.lastResponse.headers.${params.counter}`, 0)
	),

	// Calculate the value of param. This hoc is polymorphic.
	polymorphic = (params: {
		param: mixed,
		props: mixed,
		state: Object,
	}) => {
		if(typeof params.param === 'function')
			return params.param(params.props);

		if(typeof params.param === 'string')
			return _.get(params.state, params.param);

		return params.param;
	};

export default (hocParams: {
	baseURL?: string,
	entity: string | (props: Object) => string,
	fetchOnMount?: boolean,
	filters: string | Array<TFilterItem>,
	key: string | (props: Object) => string,
	page: Function | string | number,
	paramsChangeCallback?: Function,
	queryParams?: (props: Object) => Object,
	sorting: Function | string | {column: string, order: string},
	urlParams?: (props: Object) => Object,
}) => (WrappedComponent: any) => compose(
	base(hocParams),

	connect(
		(state, props: {
			faFetched: Object,
			keyStatePath: string,
			resolvedKey: string,
		}) => ({
			faFetched: {
				...props.faFetched,

				[props.resolvedKey]: {
					...props.faFetched[props.resolvedKey],
					filters   : polymorphic({param: hocParams.filters, props, state}),
					page      : polymorphic({param: hocParams.page, props, state}) || 1,
					pagesCount: getCounter({counter: 'pages-count', keyStatePath: props.keyStatePath, state}),
					sorting   : polymorphic({param: hocParams.sorting, props, state}),
					totalCount: getCounter({counter: 'total-count', keyStatePath: props.keyStatePath, state}),
				},
			},
		}),
	),

	withHandlers({

		// GET/POST params
		getQueryParams: props => nextProps => {
			const

				/* eslint-disable sort-vars */
				activeProps = (nextProps || props).faFetched[props.resolvedKey],

				filters = _.fromPairs((activeProps.filters || []).map(
					F => ([_.compact([F.field, F.operator]).join('-'), F.compiledValue || F.value])
				)),

				/* eslint-enable sort-vars */
				sorting = _.get(activeProps, 'sorting.column', '');

			let
				request = {};

			if(!_.isEmpty(filters))
				request.filterParams = JSON.stringify(filters);

			if(sorting)
				request.sort = JSON.stringify({[sorting]: _.get(activeProps, 'sorting.order') === 'asc' ? 1 : -1});

			if(activeProps.page)
				request.page = activeProps.page;

			return _.extend(request, activeProps.queryParams);
		},
	}),

	lifecycle({
		componentWillReceiveProps(nextProps) {
			// WARN do not sync if page changed from undefined to 1
			if(nextProps.hasPropsChanged({
				nextProps,
				props     : this.props,
				trackProps: ['filters', 'page', 'sorting', 'queryParams', 'urlParams'],
			})) {
				// WARN Reset sync flag due redux-api cache request
				this.props.faFetched[this.props.resolvedKey].reset('sync');

				this.props.faFetched[this.props.resolvedKey].sync(

					// URL params and query string params must be merged into single object
					// redux-api is able to distinguish URL template params and query string params
					_.extend(
						{},
						nextProps.faFetched[nextProps.resolvedKey].urlParams,
						nextProps.getQueryParams(nextProps),
					),

					// If fetch params changed — call params change callback, ignore page changes
					() => hocParams.paramsChangeCallback &&

						nextProps.hasPropsChanged({
							nextProps,
							props     : this.props,
							trackProps: ['filters', 'sorting', 'queryParams', 'urlParams'],
						}) &&

						!nextProps.hasPropsChanged({nextProps, props: this.props, trackProps: ['page']}) &&
						hocParams.paramsChangeCallback(this.props.dispatch),
				);
			}
		},
	}),

	syncOnMount(hocParams.fetchOnMount),
)(WrappedComponent);
